/**
* @file    TargetAcquisition.c
* @author  Philipp Löschl & Roland Ottensamer (roland.ottensamer@univie.ac.at)
* @date    July, 2017
*
* @defgroup TargetAcquisition Target Acquisition
* @ingroup EngineeringAlgorithms
*
* @brief Target Acquisition serves as the control center for the @StarExtraction
*        and the identification algorithms @AngleMethod, @MagnitudeValidation.
*
* ## Introduction
* The Target Acquisition is the entirety of the @StarExtraction "Star Extraction",
* @AngleMethod "Angle Method Algorithm" and
* @MagnitudeValidation "Magnitude Validation Algorithms" and is mainly responsible
* for their configuration and execution. It is controlled by an operator in
* the ground segment via the Star Map Command, which is an observation specific
* information package that provides all the necessary parameters for an
* autonomous and successful target identification. In addition, it also serves
* as the connection to the rest of the Instrument Application Software (IASW)
* and therefore also handles the incoming image data and outgoing centroiding
* reports.
*
* The acquisition process starts with the memory allocation for all
* acquisition algorithm specific variables. Once they are in place it receives
* the image data of the current observation and the respective acquisition
* algorithm parameters from the Star Map Command. This information is then used
* to scale the time dependant quantities such as the __reference target signal__,
* __background signal__ and __detection thresholds__ to the respective
* exposure time of the current situation.
*
* After this preparation, each acquisition process continues as follows:
*
*  - Star Extraction
*  - Distortion Correction
*  - Identification Algorithm
*
* The @Star Extraction "Star Extraction" is the first step for each acquisition
* and is independent of which target identification method is used in succession.
* Since the optical system distorts the star positions on the CCD enough so
* that they can become unrecogniseable by the @AngleMethod "Angle Method" a
* distortion correction (see section below (REF) for all extracted positions
* is necessary. Given a successful star extraction and the acquisition process
* is concluded with the execution of one of the identification algorithms.
* Their selection is observation specific and described in more detail in
* section Algorithm Selection (REF) below.
* \n
* \n
* ## Star Map Command
* All the required input parameters for the acquisition algorithms are provided
* by the Star Map Command telemetry packages. They are created and tested
* for every observation by an operator of the Science Operations Center (SOC)
* with the StarMap Generator Tool. Their content can be seen in Tab.: (REF)
* below:
* <table>
* <tr> <td colspan="4" align="center"> __Star Map Telecommand__
* <tr> <th> Size [Bit] <th> Data Type <th> Parameter <th> Description
* <tr> <td> 32   <td> UINT   <td> OBSID                 <td> Unique identifier for the observation
* <tr> <td> 16   <td> USHORT <td> AcqAlgoId             <td>
* <tr> <td> 32   <td> UINT   <td> SPARE                 <td> Former parameter “TargetStar” has been deleted
* <tr> <td> 2208 <td>        <td> SkyPattern            <td> Algorithm Data – details following below
* <tr> <th colspan="4" align="center"> The SkyPattern is made up of the following data structure
* <tr> <td> 24   <td> UINT24 <td> Target_Location_X     <td> "Position on the CCD (in centipixels where the target is intended to be located. Origin = bottom left pixel corner (centre of bottom left pixel is thus [50/50]), counting in Cartesian coordinates."
* <tr> <td> 24   <td> UINT24 <td> Target_Location_Y     <td> same for Y axis
* <tr> <td> 8    <td> UCHAR  <td> Algorithm ID          <td> which algorithm is to be selected for the Target Acquisition 0… None, 1… MVA, 2… AMA, 3...AMA + MVA
* <tr> <td> 16   <td> USHORT <td> SPARE                 <td> former execution time (moved)
* <tr> <td> 16   <td> USHORT <td> SPARE                 <td> former delay time (moved)
* <tr> <td> 16   <td> USHORT <td> Distance Threshold    <td> Absolute distance between measured position and Target_Location under which we switch from acquisition to centroiding.
* <tr> <td> 8    <td> UCHAR  <td> Iterations            <td> Maximum number of iterations. 0 = inf
* <tr> <td>      <td>        <th> Algorithm Paramters   <th> Parameters needed for the tuning of the algorithm – these are generated by the StarMap Generator
* <tr> <td> 16   <td> USHORT <td> SPARE                 <td> former Rebinning Factor. Now hard coded.
* <tr> <td> 16   <td> USHORT <td> Detection Threshold   <td> Sets the detection threshold for the Star Extraction. Default: 5200 [10 ADU]
* <tr> <td> 16   <td> USHORT <td> SE_Reduced_Extraction <td> Limits the Star Extraction to a radius around the expeced Target Location. Default: False
* <tr> <td> 16   <td> USHORT <td> SE_Reduced_Radius     <td> Sets the reduced extraction radius for the Star Extraction. Default: 392 pix
* <tr> <td> 16   <td> USHORT <td> SE_Tolerance          <td> Maximum positional error in Star Extraction results. Default: 5 pix
* <tr> <td> 16   <td> USHORT <td> MVA_Tolerance         <td> Maximum photometric error for the Magnitude Validation Method. Default: 15 percent
* <tr> <td> 16   <td> USHORT <td> AMA_Tolerance         <td> Maximum error for distance vectors. Default: 3 pix
* <tr> <td> 16   <td> USHORT <td> Pointing_Uncertainty  <td> Maximum expected pointing error from the Target Location. Default: 150 pix
* <tr> <td> 32   <td> UINT   <td> Target Signal         <td> Signal of the target in ADU/s
* <tr> <td> 16   <td> USHORT <td> Number of Stars       <td> Number of stars included in the sky pattern
* <tr> <td> 1920 <td> BINARY <td> Pattern               <td> Data field („Sky Pattern“) with star entries according to algorithm 920 bits = 120 shorts
* </Table>
* \n
* \n
* ## Distortion Correction
* The optical system of CHEOPS introduces significant distortion effects in the
* outer regions of the FOV. As they are strong enough to affect the intra-star
* distances that are required by the Angle Method a distortion correction is used
* to transform the Star Extraction provided positions into undistorted positions
* that match the reference stars. More information is provided by (REF).
* \n
* \n
* ## Algorithm Selection
* The acquisition algorithm is selected by the operator during the parameter
* setup for the Star Map Command. Depending on the star density and distribution
* in a specific FOV there are three possible options:
*
*  - Magnitude Validation Acquisition Algorithm (MVAA)
*  - Angle Method Algorithm (AMA)
*  - Angle Method Algorithm + Magnitude Validation Algorithm (AMA + MVA)
*
* Magnitude Validation Acquisition is meant for bright objects that can easily
* be distinguished from the other stars in an observation. The Star Map Generator
* Tool is currently configured to prefer this algorithm for targets of up to
* 10mag. Fainter stars might have another, similarly bright object in the FOV
* and could therefore be misidentified.
*
* For this reason the two Angle Method Algorithm variations are used for
* target identifications of >10mag stars. The third method is prefered and will
* be used for every case that has available brightness data of the target star.
* Here the Magnitude Validation Algorithm is used to as a backup identification
* method to verify observations with poor AMA results. Observations without this
* data will solely use the Angle Method Algorithm.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "CrIaDataPool.h"
#include "CrIaDataPoolId.h"

#include "AngleMethod.h"
#include "TaDatatypes.h"
#include "StarExtractor.h"
#include "TargetAcquisition.h"
#include "StarPosDistortCorr.h"
#include "EngineeringAlgorithms.h"

#include "SdpBuffers.h"
#include "SdpAlgorithmsImplementation.h"

struct match_candidates_t *g_candidates;
struct unique_output_t *g_output;
struct unique_output_t *g_unique_base0;
struct unique_output_t *g_unique_stars;
struct unique_output_t *g_unique_duplicate_entry;
struct where_output_t *g_multi_base_skip_index;
struct where_output_t *g_multi_base_index;
struct where_output_t *g_duplicate_base0;
struct where_output_t *g_duplicate_base1;
struct unique_output_t *g_unique_stars;
struct where_output_t *g_base0_count;
struct img_t *g_img;
struct obs_pos_t *obs_pos;
struct ref_pos_t *ref_pos;
struct ref_db_t *ref_DB;
struct obs_db_t *obs_DB;
struct match_results_t *amaresults;
struct mva_results_t *mvaresults;
struct ROI_t *g_POI;
struct ROI_t *temp_POI;

struct imgBin_t *g_imgBin;

#ifdef GROUNDSW
      # define ADVGROUNDSW 1
#endif

#define MAX_TASTARS MAXNROFSTARS

#define ALIGNSIZE(s) ((s+7) & 0xFFFFFFF8)

void initMatchingBuffers (unsigned int *data)
{
  g_img      = (struct img_t *)SWAPALLOC(ALIGNSIZE(sizeof(struct img_t)), SWAP);                     /* shall point to the unpacked full frame of 1024x1024 32 bit in CIB */
  ref_pos    = (struct ref_pos_t *)SWAPALLOC(ALIGNSIZE(sizeof(struct ref_pos_t)), SWAP);             /* < 1 kiB */
  obs_pos    = (struct obs_pos_t *)SWAPALLOC(ALIGNSIZE(sizeof(struct obs_pos_t)), SWAP);             /* < 1.5 kiB*/
  ref_DB     = (struct ref_db_t *)SWAPALLOC(ALIGNSIZE(sizeof(struct ref_db_t)), SWAP);               /* < 40 kiB */
  obs_DB     = (struct obs_db_t *)SWAPALLOC(ALIGNSIZE(sizeof(struct obs_db_t)), SWAP);               /* ca 3.2 MiB */
  amaresults = (struct match_results_t *)SWAPALLOC(ALIGNSIZE(sizeof(struct match_results_t)), SWAP); /* ca 600 kiB */
  mvaresults = (struct mva_results_t *)SWAPALLOC(ALIGNSIZE(sizeof(struct mva_results_t)), SWAP);     /* ca 132 bytes */

  g_candidates             = (struct match_candidates_t *)AUXALLOC(ALIGNSIZE(sizeof(struct match_candidates_t)), AUX); /* ca 416 kB */
  g_output                 = (struct unique_output_t *)AUXALLOC(ALIGNSIZE(sizeof(struct unique_output_t)), AUX);       /* 480 B */
  g_unique_base0      = (struct unique_output_t *)AUXALLOC(ALIGNSIZE(sizeof(struct unique_output_t)), AUX);       /* 480 B */
  g_unique_duplicate_entry = (struct unique_output_t *)AUXALLOC(ALIGNSIZE(sizeof(struct unique_output_t)), AUX);       /* 480 B */
  g_multi_base_skip_index  = (struct where_output_t *)AUXALLOC(ALIGNSIZE(sizeof(struct where_output_t)), AUX);         /* 240 B */
  g_multi_base_index       = (struct where_output_t *)AUXALLOC(ALIGNSIZE(sizeof(struct where_output_t)), AUX);         /* 240 B */
  g_duplicate_base0        = (struct where_output_t *)AUXALLOC(ALIGNSIZE(sizeof(struct where_output_t)), AUX);         /* 240 B */
  g_duplicate_base1        = (struct where_output_t *)AUXALLOC(ALIGNSIZE(sizeof(struct where_output_t)), AUX);         /* 240 B */
  g_unique_stars           = (struct unique_output_t *)AUXALLOC(ALIGNSIZE(sizeof(struct unique_output_t)), AUX);       /* 480 B */
  g_base0_count            = (struct where_output_t *)AUXALLOC(ALIGNSIZE(sizeof(struct where_output_t)), AUX);         /* 240 B */

  g_imgBin = (struct imgBin_t *) AUXALLOC(ALIGNSIZE(sizeof(struct imgBin_t)), AUX); /* 65538 B */
  g_POI    = (struct ROI_t *) AUXALLOC(ALIGNSIZE(sizeof(struct ROI_t)), AUX);       /* 131078 B */
  temp_POI = (struct ROI_t *) AUXALLOC(ALIGNSIZE(sizeof(struct ROI_t)), AUX);       /* 131078 B */ /* Mantis 2268 */

  g_img->data = data;
  g_img->xdim = TA_XDIM;
  g_img->ydim = TA_YDIM;

  return;
}


void free_matching_buffers (void)
{
#if (__sparc__)
  release(AUX);
  release(SWAP);
#else
  swapalloc(0,1);
  auxalloc(0,1);
#endif

  return;
}


void TargetAcquisition (struct CrIaSib *Sib)
{
  float TargetLocation[2];
  float ExposureTime;

  int SignalThrd, CandidateRadius, Shift[2];
  unsigned int i, u32temp, TargetSignal, DetectionThld;
  unsigned char u8temp, TaAlgorithmId;

  unsigned short u16temp, RebinFactor, DetTh;
  unsigned short ReducedExtraction, ExtractionRadius, ExtractionTolerance;
  unsigned short MvaTolerance, AmaTolerance;
  unsigned short PointUncert, BgSignal, NrOfStars;

  int xstart;

  unsigned short cibIn;
  unsigned int *cibAddress;

  unsigned char distCorr;
  float optAxisX, optAxisY;

  float TaMaxSigFract, TaBias, TaDark, TaSkyBg;
  
  struct SibHeader *sibHeader;

  /*
     copy the image from expos to CIB and crop to 1024 x 1024
  */

  /* determine nominal or redundant side to find the proper image offset in X */
  sibHeader = (struct SibHeader *) Sib->Header;

  xstart = LSM_XDIM; /* NOM: 28 */
  
  if (GetFpmSide(sibHeader->CcdTimingScriptId) == FPM_SIDE_B)
    {      
      xstart = RSM_XDIM; /* RED: 24 */
    }

  CrIaCopy (CIBIN_ID, &cibIn);
  cibAddress = (unsigned int *)GET_ADDR_FROM_SDB_OFFSET(cibIn);

  CropCopyFromSibAndUnpack (Sib->Expos, FULL_SIZE_X, FULL_SIZE_Y, 1024, 1024, xstart, 0, cibAddress);

  /* first we release the necessary heaps */
  free_matching_buffers();

  initMatchingBuffers(cibAddress);

  /*  TA_DEBUG("first pixels are: %u %u %u\n", g_img->data[0], g_img->data[1], g_img->data[2]); */

  /*
     assign DP variables
  */

  /* obsid is not needed here */

  /* acqalgoid is used one level up */

  /* spare u32 is not used */

  /* tglocx 24 */
  CrIaCopy(TARGETLOCATIONX_ID, &u32temp);
  TargetLocation[0] = (float) ((u32temp - 50) / 100.f); /* comes in centipixel */ /* was: .. - 2850 .., but changed upon request by platform */
  /* TA_DEBUG("TARGETLOCX: %f\n", TargetLocation[0]); */

  /* tglocy 24 */
  CrIaCopy(TARGETLOCATIONY_ID, &u32temp);
  TargetLocation[1] = (float) ((u32temp - 50) / 100.f); /* comes in centipixel */
  /* TA_DEBUG("TARGETLOCY: %f\n", TargetLocation[1]); */

  /* algoid is 16 in DP but 8 in command */
  CrIaCopy(TAALGOID_ID, &u16temp);
  TaAlgorithmId = (unsigned char)u16temp;
  /* TA_DEBUG("TAALGOID: %u\n", TaAlgorithmId); */

  /* TATIMINGPAR1 is used one level up */

  /* TATIMINGPAR2 is used one level up */

  /* TADISTANCETHRD is used one level up */

  /* TAITERATIONS is used one level up */

  /* rebinfact ushort */
  CrIaCopy(TAREBINNINGFACT_ID, &RebinFactor);
  RebinFactor = 8; /* NOTE: must be 8! */

  /* detthrd ushort */
  CrIaCopy(TADETECTIONTHRD_ID, &DetTh);
  DetectionThld = DetTh * 10;
  /* TA_DEBUG("TADETECTIONTHRD: %u\n", DetectionThrd); */

  /* se_reduced_extr ushort */
  CrIaCopy(TASEREDUCEDEXTR_ID, &ReducedExtraction);
  /* TA_DEBUG("TASEREDUCEEXTR: %u\n", ReducedExtraction); */

  /* se reduced radius 16 */
  CrIaCopy(TASEREDUCEDRADIUS_ID, &ExtractionRadius);
  /* TA_DEBUG("TASEREDUCEDRADIUS: %u\n", ExtractionRadius); */

  /* se tolerance 16 */
  CrIaCopy(TASETOLERANCE_ID, &ExtractionTolerance);
  /* TA_DEBUG("TASETOLERANCE: %u\n", ExtractionTolerance); */

  /* mva tolerance 16 */
  CrIaCopy(TAMVATOLERANCE_ID, &MvaTolerance);
  /* TA_DEBUG("TAMVATOLERANCE: %u\n", MvaTolerance); */

  /* ama tolerance 16 */
  CrIaCopy(TAAMATOLERANCE_ID, &AmaTolerance);
  /* TA_DEBUG("TAAMATOLERANCE: %u\n", AmaTolerance); */

  /* pointing uncertainty 16 */
  CrIaCopy(TAPOINTUNCERT_ID, &PointUncert);
  /* TA_DEBUG("TAPOINTUNCERT: %u\n", PointUncert); */
  CandidateRadius = PointUncert + ExtractionTolerance;

  /*
     copy data pattern parameters
  */

  /* targetsig 32 */
  CrIaCopy(TATARGETSIG_ID, &TargetSignal);

  /* nrofstars 16 */
  CrIaCopy(TANROFSTARS_ID, &NrOfStars);
  /* TA_DEBUG("TANROFSTARS: %u\n", NrOfStars); */

  /*
     copy stars pattern
  */


  /* now read the stars */
  for (i=0; (i < NrOfStars) && (i < MAX_TASTARS); i++)
    {
      CrIaCopyArrayItem (STARMAP_ID, &u8temp, 4*i);
      u16temp = ((unsigned short)u8temp) << 8;
      CrIaCopyArrayItem (STARMAP_ID, &u8temp, 4*i+1);
      u16temp |= (unsigned short)u8temp;
      ref_pos->x[i] = (float) u16temp / 10.0f;

      CrIaCopyArrayItem (STARMAP_ID, &u8temp, 4*i+2);
      u16temp = ((unsigned short)u8temp) << 8;
      CrIaCopyArrayItem (STARMAP_ID, &u8temp, 4*i+3);
      u16temp |= (unsigned short)u8temp;
      ref_pos->y[i] = (float) u16temp / 10.0f;
    }

  ref_pos->n = i;

  /*
    calculate the background
  */

  /* get exposure time from SIB */
  ExposureTime = sibHeader->ExposureTime / 1000.0f;
  
  CrIaCopy(TABIAS_ID, &TaBias);
  CrIaCopy(TADARK_ID, &TaDark);
  CrIaCopy(TASKYBG_ID, &TaSkyBg); /* 8.35 ph/s (zody) + 2 ph/s (straylight) = 10.35 photons/s; 0.65 e/ph is the QE; 0.439 is the gain in e/ADU; see Master thesis Ferstl p 111 */
  
  BgSignal = calculate_background_signal (ExposureTime, TaBias, TaDark, TaSkyBg);

  TargetSignal = TargetSignal * ExposureTime;
  SignalThrd = (int)(DetectionThld * ExposureTime);

  CrIaCopy (TAMAXSIGFRACT_ID, &TaMaxSigFract);
  DetectionThld = (int)(SignalThrd * TaMaxSigFract);


#ifdef GROUNDSW
  TA_DEBUG("Exptime       : %f\n", ExposureTime);
  TA_DEBUG("BgSignal      : %u\n", BgSignal);
  TA_DEBUG("TargetSignal  : %u\n", TargetSignal);
  TA_DEBUG("SignalThrd    : %d\n", SignalThrd);
  TA_DEBUG("DetectionThld : %u\n", DetectionThld);
#endif

  star_extraction (g_img, obs_pos, RebinFactor, TargetLocation, ReducedExtraction, ExtractionRadius, DetectionThld, SignalThrd, BgSignal, g_POI);
  
  /* carry out distortion correction */
  CrIaCopy (DIST_CORR_ID, &distCorr);
  
  if (distCorr  != 0)
    {
      CrIaCopy(OPT_AXIS_X_ID, &optAxisX);
      CrIaCopy(OPT_AXIS_Y_ID, &optAxisY);
      
      (void) star_pos_distort_corr (obs_pos, optAxisX, optAxisY, TA_XDIM, TA_YDIM);
    }

  Shift[0] = 0;
  Shift[1] = 0;

  /* start with a dummy report */
  PrepareSibCentroid (Sib, CEN_INV_TARGET, 0.0f, 0.0f);

  /* NOTE: Depending on the application either the distortion corrected obs_pos
           or the original uncorrected obs_pos have to be used */
  switch (TaAlgorithmId)
  {
    case TA_ALGO_NONE :

      /* no matching, use dummy report */

      break;

    case TA_ALGO_MVA :

      if ((obs_pos->n > 0) && (TargetSignal > 0))
      {
        /* use uncorrected obs_pos to measure the signal on the right coordinates */
        target_magntitude_validation_acquisition (mvaresults, obs_pos, TargetLocation, CandidateRadius, g_img, TargetSignal, MvaTolerance, BgSignal);

        if (mvaresults->valid_signal != 0)
        {
          /* use uncorrected obs_pos to measure the correct offset on the CCD*/
          calculate_target_shift_MVA (obs_pos, mvaresults, Shift, TargetLocation);
#ifdef GROUNDSW
          TA_DEBUG("MVA says: %d, %d centi-pixel\n", Shift[0], Shift[1]);
#endif
          PrepareSibCentroid (Sib, CEN_VAL_FULL, Shift[0]/100.0f, Shift[1]/100.0f);
        }
#ifdef ADVGROUNDSW
        else
	  printf("MVA: %f ", 0.0);
#endif
      }

      break;

    case TA_ALGO_AMA :

      if ((obs_pos->n > 3) && (ref_pos->n > 3))
      {
        create_ref_DB (ref_pos, 0, ref_DB, AmaTolerance); /* NOTE: target index is always 0 */

        /* use corrected obs_pos to match the reference fingerprints */
        create_obs_DB (obs_pos, obs_DB, TargetLocation, CandidateRadius, AmaTolerance);
	
        match_databases (ref_DB, obs_DB, amaresults);
	
        /* both obs_pos possible, since only used GROUNDSW outputs */
	determine_match_quality (amaresults, obs_pos);
	
        if (amaresults->valid_quality != 0)
        {
          /* use uncorrected obs_pos to measure the correct offset on the CCD*/
          calculate_target_shift_AMA (obs_pos, amaresults, Shift, TargetLocation);
#ifdef GROUNDSW
          TA_DEBUG("AMA says: %d, %d centi-pixel\n", Shift[0], Shift[1]);
#endif
          PrepareSibCentroid (Sib, CEN_VAL_FULL, Shift[0]/100.0f, Shift[1]/100.0f);
        }
        else
        {
#ifdef ADVGROUNDSW
	  /*      printf("AMA: %f ", amaresults->match_quality); */
      /*printf("TACQ: FAILED ");*/
#endif
          /* ERROR: MATCH QUALITY < 50% of DB -> TACQ FAIL */
          PrepareSibCentroid (Sib, CEN_INV_TARGET, 0.0f, 0.0f);
        }
      }
      else
      {
#ifdef ADVGROUNDSW
      /*printf("TACQ: FAILED ");*/
	/*      printf("AMA: %f ", 0.0); */
#endif
        /* ERROR: NOT ENOUGH STARS FOR AMA */
        PrepareSibCentroid (Sib, CEN_INV_INPUT, 0.0f, 0.0f);
      }
      break;

    case TA_ALGO_AMAMVA :

      if ((obs_pos->n > 3) && (ref_pos->n > 3))
      {
        create_ref_DB (ref_pos, 0, ref_DB, AmaTolerance); /* NOTE: target index is always 0 */

        /* use corrected obs_pos to match the reference fingerprints */
        create_obs_DB (obs_pos, obs_DB, TargetLocation, CandidateRadius, AmaTolerance);

        match_databases (ref_DB, obs_DB, amaresults);

        /* both obs_pos possible, since only used GROUNDSW outputs */
	determine_match_quality (amaresults, obs_pos);

        if (amaresults->valid_quality != 0)
        {
          /* use uncorrected obs_pos to measure the correct offset on the CCD*/
          calculate_target_shift_AMA (obs_pos, amaresults, Shift, TargetLocation);
#ifdef ADVGROUNDSW
	  /*          printf("AMA: %f ", amaresults->match_quality); */
#endif
          PrepareSibCentroid (Sib, CEN_VAL_FULL, Shift[0]/100.0f, Shift[1]/100.0f);
        }
        else if ((amaresults->match_quality > 40) && (TargetSignal > 0)) /* 40% of DB */
        {
          /* use uncorrected obs_pos to measure the signal on the right coordinates */
          target_magnitude_validation (g_img, obs_pos, amaresults, BgSignal, TargetSignal, MvaTolerance);
          if (amaresults->valid_signal != 0)
          {
            /* insuff matches but MVA rescue*/
            /* use uncorrected obs_pos to measure the correct offset on the CCD*/
            calculate_target_shift_AMA (obs_pos, amaresults, Shift, TargetLocation);
#ifdef ADVGROUNDSW
	    /*      printf("AMA: %f ", amaresults->match_quality); */
#endif
	    PrepareSibCentroid (Sib, CEN_VAL_FULL, Shift[0]/100.0f, Shift[1]/100.0f);
          }
          else
          {
            /* negative insuff matches and no MVA rescue */
#ifdef ADVGROUNDSW
	    /*      printf("AMA: %f ", amaresults->match_quality); */
      /*printf("TACQ: FAILED ");*/
#endif
            PrepareSibCentroid (Sib, CEN_INV_SMEAR, 0.0f, 0.0f);
          }
        }
        else
        {
#ifdef ADVGROUNDSW
	  /*      printf("AMA: %f ", amaresults->match_quality); */
      /*printf("TACQ: FAILED ");*/
#endif
          /* negative insuff matches */
          PrepareSibCentroid (Sib, CEN_INV_SMEAR, 0.0f, 0.0f);
        }
      }
      else
      {
#ifdef ADVGROUNDSW
      /*printf("TACQ: FAILED ");*/
	/*      printf("AMA: %f ", 0.0); */
#endif
        	/* ERROR: NOT ENOUGH STARS FOR AMA */
        	PrepareSibCentroid (Sib, CEN_INV_INPUT, 0.0f, 0.0f);
      }

      break;

    default:

#ifdef ADVGROUNDSW
      /*printf("TACQ: FAILED ");*/
      /*      printf("AMA: %f ", 0.0); */
#endif
      /* invalid algorithm id */
      PrepareSibCentroid (Sib, CEN_INV_ALGO, 0.0f, 0.0f);

      break;

    }

  free_matching_buffers();

  return;
}
